import numpy as np
import math
from soio.core.problem import FloatProblem
from soio.core.solution import FloatSolution

class Ackley(FloatProblem):
    '''
    https://www.sfu.ca/~ssurjano/ackley.html
    Global minimum: f(X) = 0 at X = [0]D
    '''
    def __init__(self, D, a=20, b=0.2, c=2*math.pi):
        self.number_of_variables = D
        self.lower_bound = np.ones(D)*(-32.768)
        self.upper_bound = np.ones(D)*32.768
        self.a = a
        self.b = b
        self.c = c
        self.nfes = 0

    def get_name(self):
        return 'Ackley Function'

    def evaluate(self, solution: FloatSolution):
        self.nfes += 1
        solution.objective = self.compute(solution.variables)

    def compute(self, X: np.ndarray):
        return -self.a * math.exp(-self.b * math.sqrt(np.mean(np.square(X)))) \
               - math.exp(np.mean(np.cos(self.c * X))) + self.a + math.e

    def create_solution(self) -> FloatSolution:
        new_solution = FloatSolution(self.number_of_variables)
        new_solution.variables = np.random.uniform(self.lower_bound, self.upper_bound)
        return new_solution




'''
Shuffled Cellluar Evolutionary Grey Wolf Optimizer
@author: zzw
@log: 2019.6.4, create
'''
from soio.core.algorithm import Algorithm
from soio.core.solution import Solution
from soio.core.problem import PermutationProblem
from soio.core.operator import Operator
import time
import numpy as np
import random

'''
Cellular Automata (2D)
'''
CA_L5 = [[0,-1], [0,1], [-1,0], [1,0]]
CA_L9 = [[0,-2], [0,-1], [0,1], [0,2], [-2,0], [-1,0], [2,0], [1,0]]
CA_C9 = [[-1,-1], [0,-1], [1,-1], [-1,0], [1,0], [-1,1], [0,1], [1,1]]
CA_C13 = [[0,-2], [-1,-1], [0,-1], [1,-1], [-2,0], [-1,0], [1,0], [2,0],
          [-1,1], [0,1], [1,1], [0, 2]]
CA_C21 = [[-1,-2], [0,-2], [1,-2],
          [-2,-1], [-1,-1], [0,-1], [1,-1], [2,-1],
          [-2,0], [-1,0], [1,0], [2,0],
          [-2,1], [-1,1], [0,1], [1,1], [2,1],
          [-1,2], [0,2], [1,2]]
CA_C25 = [[-2,-2], [-1,-2], [0,-2], [1,-2], [2,-2],
          [-2,-1], [-1,-1], [0,-1], [1,-1], [2,-1],
          [-2,0], [-1,0], [1,0], [2,0],
          [-2,1], [-1,1], [0,1], [1,1], [2,1],
          [-2,2], [-1,2], [0,2], [1,2], [2,2]]

def find_neighborhood(neighbor_structure, grid_catalog, coord):
    '''
    :param neighbor_structure:  2D list, Cellular Automata
    :param grid_catalog: 2D ndarray, wolf index in each celluar grid
    :param coord:  [x,y]
    :return:  neighborhood wolves indexes
    '''
    l,w = grid_catalog.shape
    neighbors = []
    for nc in neighbor_structure:
        nx = coord[0] + nc[0]
        ny = coord[1] + nc[1]

        if nx >= l:
            nx -= l
        elif nx < 0:
            nx += l

        if ny >= w:
            ny -= w
        elif ny < 0:
            ny += w

        neighbors.append(grid_catalog[nx,ny])
    return neighbors


class CellularGreyWolfOptimizer(Algorithm):
    def __init__(self,
                 problem: PermutationProblem,
                 mutation: Operator,
                 crossover: Operator,
                 max_nfes: int,
                 grid_shape = [20,20],
                 neighbor_structure = CA_C21,
                 ):

        self.problem = problem
        self.swarm_size = grid_shape[0]*grid_shape[1]
        self.max_nfes = max_nfes

        self.mutation_operator = mutation
        self.crossover_operator = crossover

        self.neighbor_structure = neighbor_structure
        self.grid_catalog = np.random.permutation(self.swarm_size).reshape(grid_shape)

        self.wolves_coords = [None]*self.swarm_size
        for i in range(grid_shape[0]):
            for j in range(grid_shape[1]):
                self.wolves_coords[self.grid_catalog[i][j]] = [i,j]

        self.best_solution = Solution(self.problem.number_of_variables)
        self.best_solution.objective = float("inf")

        self.grid_shape = grid_shape


    def create_initial_swarm(self) -> np.ndarray:
        """ Creates the initial list of solutions of a metaheuristic. """
        swarm = [self.problem.create_solution() for _ in range(self.swarm_size)]
        return swarm


    def shuffle_swarm(self):
        # shuffle
        self.grid_catalog = np.random.permutation(self.swarm_size).reshape(self.grid_shape)

        self.wolves_coords = [None] * self.swarm_size
        for i in range(self.grid_shape[0]):
            for j in range(self.grid_shape[1]):
                self.wolves_coords[self.grid_catalog[i][j]] = [i, j]

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        self.swarm = self.create_initial_swarm()
        best_i = None
        for i in range(self.swarm_size):
            self.problem.evaluate(self.swarm[i])
            if self.swarm[i].objective < self.best_solution.objective:
                self.best_solution = self.swarm[i]

        self.records = []
        terminate = False
        while True:
            if terminate:
                break

            update = False

            for i in range(self.swarm_size):
                if self.max_nfes < self.problem.nfes:
                    terminate = True
                    break
                wolf = self.swarm[i]
                new_wolf = None
                coord = self.wolves_coords[i]
                neighbors = find_neighborhood(self.neighbor_structure, self.grid_catalog, coord)
                alpha_wolf, beta_wolf, delta_wolf  = self.select_best_three_neigbors(neighbors)
                # social hierarchy
                new_wolf = None
                if wolf.objective == self.best_solution.objective:
                    new_wolf = self.mutation_operator.execute(wolf)
                elif wolf.objective < alpha_wolf.objective:
                    new_wolf = self.crossover_operator.execute([self.best_solution, wolf])
                elif wolf.objective < beta_wolf.objective:
                    new_wolf = self.crossover_operator.execute([alpha_wolf, wolf])
                elif wolf.objective < delta_wolf.objective:
                    if random.random() < 0.5:
                        new_wolf = self.crossover_operator.execute([alpha_wolf, wolf])
                    else:
                        new_wolf = self.crossover_operator.execute([beta_wolf, wolf])
                else:
                    r = random.random()
                    if r < 0.33:
                        new_wolf = self.crossover_operator.execute([alpha_wolf, wolf])
                    elif r < 0.66:
                        new_wolf = self.crossover_operator.execute([beta_wolf, wolf])
                    else:
                        new_wolf = self.crossover_operator.execute([delta_wolf, wolf])

                    new_wolf = self.mutation_operator.execute(new_wolf)

                    if new_wolf.objective == None:
                        self.problem.evaluate(new_wolf)
                    else:
                        self.problem.nfes += 1
                    self.swarm[i] = new_wolf
                    # # greedy
                    # if new_wolf.objective < self.swarm[i].objective:
                    #     self.swarm[i] = new_wolf

                if new_wolf.objective < self.best_solution.objective:
                    self.best_solution = new_wolf
                    update = True

            if not update:
                self.shuffle_swarm()

            self.records.append(self.best_solution.objective)

        self.total_computing_time = time.time() - start_computing_time

    def select_best_three_neigbors(self, neighbors):
        alpha_wolf = Solution(self.problem.number_of_variables)
        alpha_wolf.objective = float("inf")
        beta_wolf = Solution(self.problem.number_of_variables)
        beta_wolf.objective = float("inf")
        delta_wolf = Solution(self.problem.number_of_variables)
        delta_wolf.objective = float("inf")
        for i in neighbors:
            if self.swarm[i].objective < alpha_wolf.objective:
                delta_wolf = beta_wolf
                beta_wolf = alpha_wolf
                alpha_wolf = self.swarm[i]

            elif self.swarm[i].objective < beta_wolf.objective:
                delta_wolf = beta_wolf
                beta_wolf = self.swarm[i]

            elif self.swarm[i].objective < delta_wolf.objective:
                delta_wolf = self.swarm[i]

        return alpha_wolf, beta_wolf, delta_wolf



    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Shuffled Cellular Evolutionary Grey Wolf Optimizer'



'''
Biogeography-Based Optimization
@author: zzw
@reference: Simon, D. 2008.
                    Biogeography-Based Optimization.
                    IEEE Transactions on Evolutionary Computation 12 (6):702-13.
                    doi: https://doi.org/10.1109/TEVC.2008.919004.
                    http://embeddedlab.csuohio.edu/BBO/
@log: 2019.7.25, create, still need improve
'''
from soio.core.algorithm import Algorithm
from soio.core.problem import FloatProblem
from soio.core.solution import FloatSolution
import time
import random

class BBO(Algorithm):
    def __init__(self,
        problem: FloatProblem,
        habitats_size: int,
        max_nfes: int,
    ):
        self.problem = problem
        self.habitats_size=habitats_size
        self.max_iterations = int(max_nfes/habitats_size)
        pass

    def run(self):
        """ Execute the algorithm. """
        self.records =[]
        start_computing_time = time.time()

        self.habitats = [self.problem.create_solution() for _ in range(self.habitats_size)]
        for habitat in self.habitats:
            self.problem.evaluate(habitat)
        self.habitats.sort(key=lambda x:x.objective)
        self.best_solution = self.habitats[0]
        assert self.habitats[0].objective < self.habitats[-1].objective

        self.immigrate_rates = []
        self.emigrate_rates = []
        for i in range(self.habitats_size):
            mu = (self.habitats_size+1-i)/(self.habitats_size+1)
            self.emigrate_rates.append(mu) #decreasing
            self.immigrate_rates.append(1-mu) #increasing
        self.mut_prob = 0.01
        self.keep=2

        for _  in range(self.max_iterations):
            # Migration
            new_habitats =[]
            new_habitats.append(self.habitats[0])
            for hi in range(1,self.habitats_size):
                new_habitat = FloatSolution(self.problem.number_of_variables)
                for vi in range(self.problem.number_of_variables):
                    if random.random() < self.immigrate_rates[hi]:
                        # Roulette Wheel Selection
                        rhi = self.roulette_wheel_selection()
                        new_habitat.variables[vi] = self.habitats[rhi].variables[vi]
                    else:
                        new_habitat.variables[vi] = self.habitats[hi].variables[vi]
                new_habitats.append(new_habitat)

            # Mutation
            for hi in range(self.habitats_size):
                for vi in range(self.problem.number_of_variables):
                    if random.random() < self.mut_prob:
                        new_habitats[hi].variables[vi] = self.problem.lower_bound[vi] + \
                                               (self.problem.upper_bound[vi]-self.problem.lower_bound[vi])*random.random()

            for habitat in new_habitats:
                self.problem.evaluate(habitat)

            new_habitats.sort(key=lambda x:x.objective)
            new_habitats[self.habitats_size-self.keep:] = self.habitats[:self.keep]
            self.habitats = sorted(new_habitats, key=lambda x:x.objective)
            assert len(self.habitats)==self.habitats_size

            self.best_solution = self.habitats[0]
            self.records.append(self.best_solution.objective)

        self.total_computing_time = time.time() - start_computing_time



    def roulette_wheel_selection(self):
        r = random.random()*sum(self.emigrate_rates)
        s = self.emigrate_rates[0]
        si = 0
        while (r>s) and (si<(self.habitats_size-1)):
            si += 1
            s += self.emigrate_rates[si]
        return si



    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Biogeography-Based Optimization'


from abc import  abstractmethod
from soio.core.solution import Solution

class Operator:
    """ Class representing operator """

    @abstractmethod
    def execute(self, Solution) ->Solution:
        pass

    @abstractmethod
    def get_name(self) -> str:
        pass


'''
Grey WOlf OptimizationAlgorithm
@author: zzw
@reference: Mirjalili, S., S. M. Mirjalili, and A. Lewis. 2014.
                    Grey Wolf Optimizer."
                    Advances in Engineering Software 69:46-61
                    doi: 10.1016/j.advengsoft.2013.12.007.

@log: 2019.5.12, create
'''
from soio.core.algorithm import Algorithm
from soio.core.solution import FloatSolution
from soio.core.problem import FloatProblem
import time
import numpy as np
import copy

class GreyWolfOptimizer(Algorithm):

    def __init__(self,
                 problem: FloatProblem,
                 swarm_size: int,
                 max_nfes: int):

        self.problem = problem
        self.swarm_size = swarm_size
        self.max_iterations = int(max_nfes/swarm_size)

        self.alpha_wolf = FloatSolution(self.problem.number_of_variables)
        self.alpha_wolf.objective = float("inf")
        self.beta_wolf = FloatSolution(self.problem.number_of_variables)
        self.beta_wolf.objective = float("inf")
        self.delta_wolf = FloatSolution(self.problem.number_of_variables)
        self.delta_wolf.objective = float("inf")

    def create_initial_swarm(self):
        """ Creates the initial list of solutions of a metaheuristic. """
        return [self.problem.create_solution() for _ in range(self.swarm_size)]

    def evaluate(self, swarm):
        """ Evaluates the swarm. """
        for i in range(self.swarm_size):
            self.problem.evaluate(swarm[i])
            if swarm[i].objective < self.alpha_wolf.objective:
                self.delta_wolf = self.beta_wolf
                self.beta_wolf = self.alpha_wolf
                self.alpha_wolf = swarm[i]

            elif swarm[i].objective < self.beta_wolf.objective:
                self.delta_wolf = self.beta_wolf
                self.beta_wolf =  swarm[i]

            elif swarm[i].objective < self.delta_wolf.objective:
                self.delta_wolf =  swarm[i]
        return swarm

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        self.swarm = self.create_initial_swarm()
        self.evaluate(self.swarm)
        self.records = []

        for iter in range(self.max_iterations):
            a = 2 - iter * 2 / self.max_iterations  # a decreases linearly fro 2 to 0
            r1 = np.random.random((3, self.swarm_size, self.problem.number_of_variables))
            r2 = np.random.random((3, self.swarm_size, self.problem.number_of_variables))
            A1 = 2 * a * r1[0] - a
            C1 = 2 * r2[0]
            A2 = 2 * a * r1[1] - a
            C2 = 2 * r2[1]
            A3 = 2 * a * r1[2] - a
            C3 = 2 * r2[2]

            for i in range(self.swarm_size):
                wolf = self.swarm[i]
                D_alpha = np.abs(C1[i] * self.alpha_wolf.variables - wolf.variables)
                X1 = self.alpha_wolf.variables - A1[i] * D_alpha

                D_beta = np.abs(C2[i] * self.beta_wolf.variables - wolf.variables)
                X2 = self.beta_wolf.variables - A2[i] * D_beta

                D_delta = np.abs(C3[i] * self.delta_wolf.variables - wolf.variables)
                X3 = self.delta_wolf.variables - A3[i] * D_delta

                new_pos = (X1 + X2 + X3) / 3
                new_pos = np.clip(new_pos, self.problem.lower_bound, self.problem.upper_bound)

                new_wolf = FloatSolution(self.problem.number_of_variables)
                new_wolf.variables = new_pos
                self.swarm[i] = new_wolf

            self.swarm = self.evaluate(self.swarm)
            self.records.append(self.alpha_wolf.objective)

        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.alpha_wolf

    def get_name(self) -> str:
        return 'Grey Wolf Optimizer'













'''
Whale Optimization Algorithm
@author: zzw
@reference: Mirjalili, S., & Lewis, A. (2016).
                    The Whale Optimization Algorithm.
                    Advances in Engineering Software, 95:51-67.
                    doi:https://doi.org/10.1016/j.advengsoft.2016.01.008

@log: 2019.5.14, creation
'''
from soio.core.algorithm import Algorithm
from soio.core.solution import FloatSolution
from soio.core.problem import FloatProblem
import time
import numpy as np
import random
import math

class WhaleOptimizer(Algorithm):

    def __init__(self,
                 problem: FloatProblem,
                 swarm_size: int,
                 max_nfes: int):

        self.problem = problem
        self.swarm_size = swarm_size
        self.max_iterations = int(max_nfes/swarm_size)

        self.best_solution = FloatSolution(self.problem.number_of_variables)
        self.best_solution.objective = float("inf")

    def create_initial_swarm(self) -> np.ndarray:
        """ Creates the initial list of solutions of a metaheuristic. """
        return [self.problem.create_solution() for _ in range(self.swarm_size)]

    def evaluate(self, swarm):
        """ Evaluates the swarm. """
        for i in range(self.swarm_size):
            self.problem.evaluate(swarm[i])
            if swarm[i].objective < self.best_solution.objective:
                self.best_solution = swarm[i]

        return swarm

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        self.swarm = self.create_initial_swarm()
        self.swarm = self.evaluate(self.swarm)
        self.records = []

        for iter in range(self.max_iterations):
            a = 2 - iter * ((2) / self.max_iterations)  # a decreases linearly fro 2 to 0
            a2 = -1 + iter * ((-1) / self.max_iterations) #  a2 linearly decreases from -1 to -2

            for i in range(self.swarm_size):
                r1 = random.random()
                r2 = random.random()
                A = 2 * a * r1 - a
                C = 2 * r2
                b = 1
                l = (a2 - 1) * random.random() + 1

                p = random.random()
                new_whale = FloatSolution(self.problem.number_of_variables)
                for j in range(0, self.problem.number_of_variables):

                    if p < 0.5:
                        if abs(A) >= 1:
                            rand_leader_index = random.randint(0, self.swarm_size-1)
                            X_rand = self.swarm[rand_leader_index].variables
                            D_X_rand = abs(C * X_rand[j] - self.swarm[i].variables[j])
                            new_whale.variables[j] = X_rand[j] - A * D_X_rand

                        elif abs(A) < 1:
                            D_Leader = abs(C * self.best_solution.variables[j] - self.swarm[i].variables[j])
                            new_whale.variables[j] = self.best_solution.variables[j] - A * D_Leader

                    elif p >= 0.5:
                        distance2Leader = abs(self.best_solution.variables[j] - self.swarm[i].variables[j])
                        new_whale.variables[j] = distance2Leader * math.exp(b * l) * math.cos(l * 2 * math.pi) + self.best_solution.variables[j]

                new_whale.variables = np.clip(new_whale.variables, self.problem.lower_bound, self.problem.upper_bound)
                self.swarm[i]=new_whale

                # # greedy
                # if new_whale.objective < self.swarm[i].objective:
                #     self.swarm[i] = new_whale
                #     if new_whale.objective < self.best_solution.objective:
                #         self.best_solution = new_whale


            self.evaluate(self.swarm)
            self.records.append(self.best_solution.objective)

        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Whale Optimization Algorithm'




'''
Migrating Birds Optimization (MBO)
@author: zzw
@reference:
@log: 2019.6.8, create
'''
from soio.core.algorithm import Algorithm
from soio.core.problem import PermutationProblem
from soio.core.operator import Operator
import time


class MigratingBirdsOptimizer(Algorithm):
    def __init__(self,
                 problem: PermutationProblem,
                 neighbor_generator: Operator,
                 max_nfes: int,
                 swarm_size: int,
                 neighbors_size: int,
                 shared_neigbors: int,
                 num_tours: int):
        '''
        :param problem:
        :param neighbor_generator:
        :param max_nfes:
        :param swarm_size:  the number of individuals in the swarm (α), odd
        :param neighbors_size:  the number of neigbors to be considered (β), β >= (2χ+1)
        :param shared_neigbors: he number of neighbors to be shared with the following company (χ)
        :param num_tours: the number of tours (w)

        Step1. Initialization:
        At first,  α solutions are generated randomly, in which one solution is selected as the leader
        and others are divided equally to form the left line and the right line of the V-shape respectively.

        Step2. Leader improvement:
        β solutions are generated around the leader.
        If the best neighbor brings an improvement, then the leader is replaced; otherwise, the leader keeps unchanged.
        Then, the unused neighbors are sorted in ascending order according to their objective values
        and the best 2 χ are selected and divided equally to form the left and the right sharing neighbor set respectively

        Step3. Follower improvement:
        The improvement process is conducted along the lines toward the tails.
        For a solution (such as X) in the left (right) line, β − χ neighbors are created randomly.
        Then these β − χ solutions and χ solutions in the left (right) sharing neighbor set are seen as neighborhood of X and evaluated together.
        If the best neighbor has a better fitness than X, then X is replaced.
        Subsequently, the best χ unused neighbors are utilized to rebuild the left (right) sharing neighbor set.
        The procedure above is repeated until all solutions in the left and right list have been considered.
        '''

        self.problem = problem
        self.neighbor_generator = neighbor_generator
        self.max_nfes = max_nfes
        self.swarm_size = swarm_size
        self.max_nfes = max_nfes
        self.neighbors_size = neighbors_size
        self.shared_neigbors = shared_neigbors
        self.num_tours = num_tours

        self.best_solution = None

        self.leader = None
        self.left_followers = None
        self.right_followers = None
        self.num_followers= int((self.swarm_size - 1) / 2)


    def run(self):
        start_computing_time = time.time()

        self.leader = self.problem.evaluate(self.problem.create_solution())
        self.left_followers = [self.problem.evaluate(self.problem.create_solution()) for _ in range(self.num_followers)]
        self.right_followers = [self.problem.evaluate(self.problem.create_solution()) for _ in range(self.num_followers)]

        terminate = False
        LR = False
        while True:
            if terminate:
                break
            for tour in range(self.num_tours):
                right_shared_list = None
                left_shared_list = None

                if self.max_nfes < self.problem.nfes:
                    terminate = True
                    break

                # improve the leader
                leader_neigbors = [self.problem.evaluate(self.neighbor_generator.execute(self.leader)) for _ in range(self.neighbors_size)]
                leader_neigbors.sort(key=lambda x: x.objective)
                if leader_neigbors[0].objective< self.leader.objective:
                    self.leader = leader_neigbors[0]
                    if LR:
                        right_shared_list = leader_neigbors[1:self.shared_neigbors+1]
                        left_shared_list = leader_neigbors[self.shared_neigbors+1: 2*self.shared_neigbors+1]
                    else:
                        left_shared_list = leader_neigbors[1:self.shared_neigbors + 1]
                        right_shared_list = leader_neigbors[self.shared_neigbors + 1: 2 * self.shared_neigbors + 1]
                else:
                    if LR:
                        right_shared_list = leader_neigbors[0:self.shared_neigbors]
                        left_shared_list = leader_neigbors[self.shared_neigbors: 2*self.shared_neigbors]
                    else:
                        left_shared_list = leader_neigbors[0:self.shared_neigbors]
                        right_shared_list = leader_neigbors[self.shared_neigbors: 2 * self.shared_neigbors]

                #improve the followers
                for idx in range(self.num_followers):
                    if self.max_nfes < self.problem.nfes:
                        terminate = True
                        break

                    left_neigbors = [self.problem.evaluate(self.neighbor_generator.execute(self.left_followers[idx]))
                                     for _ in range(self.neighbors_size-self.shared_neigbors)] \
                                    + left_shared_list
                    left_neigbors.sort(key=lambda x: x.objective)
                    if left_neigbors[0].objective < self.left_followers[idx].objective:
                        self.left_followers[idx] = left_neigbors[0]
                        left_shared_list = left_neigbors[1:self.shared_neigbors+1]
                    else:
                        left_shared_list = left_neigbors[0:self.shared_neigbors]

                    if self.max_nfes < self.problem.nfes:
                        terminate = True
                        break

                    right_neigbors = [self.problem.evaluate(self.neighbor_generator.execute(self.right_followers[idx]))
                                     for _ in range(self.neighbors_size - self.shared_neigbors)] \
                                    + right_shared_list
                    right_neigbors.sort(key=lambda x: x.objective)
                    if right_neigbors[0].objective < self.right_followers[idx].objective:
                        self.right_followers[idx] = right_neigbors[0]
                        right_shared_list = right_neigbors[1:self.shared_neigbors + 1]
                    else:
                        right_shared_list = right_neigbors[0:self.shared_neigbors]

            # Select a new leader
            if LR:
                self.right_followers.append(self.leader)
                self.leader = self.right_followers[0]
                self.right_follower= self.right_followers[1:]
            else:
                self.left_followers.append(self.leader)
                self.leader = self.left_followers[0]
                self.left_follower = self.left_followers[1:]

            LR = not LR

        self.best_solution = self.leader
        for solution in self.left_followers:
            if solution.objective < self.best_solution.objective:
                self.best_solution = solution
        for solution in self.right_followers:
            if solution.objective < self.best_solution.objective:
                self.best_solution = solution
        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Migrating Birds Optimizer '









'''
Genetic Algorithm
@author: zzw

@log: 2019.5.9, create
'''
from soio.core.algorithm import Algorithm
from soio.core.problem import PermutationProblem
from soio.core.operator import Operator
import time

class GeneticAlgorithm(Algorithm):

    def __init__(self,
                 problem: PermutationProblem,
                 population_size: int,
                 offspring_size: int,
                 mutation: Operator,
                 crossover: Operator,
                 selection: Operator,
                 max_nfes: int):

        self.problem = problem
        self.population_size = population_size
        self.offspring_size = offspring_size
        self.max_iterations = int(max_nfes/population_size)

        self.mutation_operator = mutation
        self.crossover_operator = crossover
        self.selection_operator = selection

        self.best_solution = None


    def create_initial_population(self):
        return [self.problem.create_solution() for _ in range(self.population_size)]

    def evaluate(self, solution_list) ->None:
        for solution in solution_list:
            self.problem.evaluate(solution)
            if self.best_solution == None:
                self.best_solution = solution
            elif self.best_solution.objective > solution.objective:
                self.best_solution = solution

    def selection(self, population):
        mating_population = []
        for i in range(self.offspring_size):
            solution = self.selection_operator.execute(population)
            mating_population.append(solution)
        return mating_population

    def reproduction(self, mating_population):

        offspring_population = []
        for i in range(0, self.offspring_size, 2):
            parents = [mating_population[i], mating_population[i +1]]
            offspring = self.crossover_operator.execute(parents)

            for solution in offspring:
                self.mutation_operator.execute(solution)
                offspring_population.append(solution)

        return offspring_population

    def replacement(self, population, offspring_population):
        population.extend(offspring_population)
        population.sort(key=lambda s: s.objective)
        return population[:self.population_size]

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        self.population = self.create_initial_population()
        self.evaluate(self.population)
        self.records = []

        for iter in range(self.max_iterations):
            mating_population = self.selection(self.population)
            offspring_population = self.reproduction(mating_population)
            self.evaluate(offspring_population)
            self.population = self.replacement(self.population, offspring_population)
            # or self.population = offspring_population
            self.records.append(self.best_solution.objective)

        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Genetic algorithm'


from abc import  abstractmethod
from soio.core.solution import Solution
import random

class Algorithm:
    """ Class representing metaheuristic algorithms. """

    def __init__(self):
        self.number_of_variables: int = 0
        self.nfes: int = 0

    @abstractmethod
    def run(self):
        """ Execute the algorithm. """
        pass

    @abstractmethod
    def get_result(self) -> Solution:
        """ :return: Optimal solution. """
        pass

    @abstractmethod
    def get_name(self) -> str:
        pass



'''
Harmony Search  Optimization Algorithm (HS)
@author: zzw
@reference: https://link.springer.com/chapter/10.1007%2F978-3-642-00185-7_1
@log: 2019.6.9, create
'''
from soio.core.algorithm import Algorithm
from soio.core.solution import FloatSolution
from soio.core.problem import FloatProblem
import time
import random

class HarmonySearch(Algorithm):

    def __init__(self,
                 problem: FloatProblem,
                 harmony_memory_size: int,
                 max_nfes: int,
                 memory_consider_rate: float = 0.9,
                 pitch_adjust_rate: float = 0.5,
                 band_width: float = 0.5,
                 ):

        self.problem = problem
        self.HMS = harmony_memory_size
        self.HMCR = memory_consider_rate
        self.PAR = pitch_adjust_rate
        self.bw = band_width
        self.max_nfes = max_nfes

        self.best_solution = None
        self.HM = None # Harmony Memory

    def get_worst_harmony_idx(self):
        worst_harmony_idx = 0
        for idx in range(1,self.HMS):
            if self.HM[idx].objective > self.HM[worst_harmony_idx].objective:
                worst_harmony_idx = idx
        return worst_harmony_idx

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()
        self.HM = [self.problem.create_solution() for _ in range(self.HMS)]
        for s in self.HM:
            self.problem.evaluate(s)
        worst_harmony_idx = self.get_worst_harmony_idx()
        terminate = False
        while True:
            if self.max_nfes < self.problem.nfes:
                terminate = True
                break
            new_harmony = FloatSolution(self.problem.number_of_variables)
            for iv in range(self.problem.number_of_variables):
                if random.random() < self.HMCR:
                    pitch = self.HM[int(random.random()*self.HMS)].variables[iv]
                    if random.random() < self.PAR:
                        pitch = pitch + (2*random.random()-1)*self.bw
                        if pitch > self.problem.upper_bound[iv]:
                            pitch = self.problem.upper_bound[iv]
                        elif pitch < self.problem.lower_bound[iv]:
                            pitch = self.problem.lower_bound[iv]
                    new_harmony.variables[iv] = pitch
                else:
                    new_harmony.variables[iv] = random.uniform(self.problem.lower_bound[iv]*1.0, self.problem.upper_bound[iv]*1.0)
            self.problem.evaluate(new_harmony)
            if new_harmony.objective < self.HM[worst_harmony_idx].objective:
                self.HM[worst_harmony_idx] = new_harmony
                worst_harmony_idx = self.get_worst_harmony_idx()

        self.best_solution = self.HM[0]
        for harmony in self.HM:
            if harmony.objective < self.best_solution.objective:
                self.best_solution = harmony

        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Harmony Search Optimizer'



from abc import  abstractmethod
from soio.core.solution import Solution
import random

class Problem:
    """ Class representing problems. """

    def __init__(self):
        self.number_of_variables: int = 0
        self.nfes: int = 0

    @abstractmethod
    def create_solution(self) -> Solution:
        """ Creates a random solution to the problem.

        :return: Solution. """
        pass

    @abstractmethod
    def evaluate(self, solution: Solution) -> Solution:
        """ Evaluate a solution. For any new problem inheriting from :class:`Problem`, this method should be
        replaced. Note that this framework ASSUMES minimization, thus solutions must be evaluated in consequence.
        assign a value to solution.objective
        :return: None. """
        self.nfes += 1
        pass

    @abstractmethod
    def get_name(self) -> str:
        pass

class PermutationProblem(Problem):
    """ Class representing permutation problems. """

    def __init__(self):
        super(PermutationProblem, self).__init__()

    def create_solution(self) -> Solution:
        pass

class FloatProblem(Problem):

    def __init__(self):
        super(FloatProblem, self).__init__()
        self.lower_bound = []
        self.upper_bound = []

    def create_solution(self) -> Solution:
        new_solution = Solution(self.number_of_variables)
        new_solution.variables = [random.uniform(self.lower_bound[i]*1.0, self.upper_bound[i]*1.0) for i in range(self.number_of_variables)]

        return new_solution


from soio.algo import BBO, CGWO, CWOA, DE,GWO, HS, ICA, JADE, WOA
from soio.benchmark.Ackley import Ackley
import matplotlib.pyplot as plt

problem = Ackley(20)

# algorithm = BBO.BBO(
#     problem = problem,
#     habitats_size = 100,
#     max_nfes = 10000,
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()

# algorithm = CGWO.CellularGreyWolfOptimizer(
#     problem = problem,
#     max_nfes = 10000,
#     grid_shape=[10, 10],
#     neighbor_structure=CGWO.CA_C9,
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()

# algorithm = CWOA.CellularWhaleOptimizer(
#     problem = problem,
#     max_nfes = 10000,
#     grid_shape=[10, 10],
#     neighbor_structure=CWOA.CA_C9,
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()

# algorithm = DE.DifferentialEvolution(
#     problem = problem,
#     max_nfes = 10000,
#     population_size=100,
#     CR = 0.5,
#     F = 0.1,
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()

# algorithm = GWO.GreyWolfOptimizer(
#     problem = problem,
#     max_nfes = 10000,
#     swarm_size=100,
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()

# algorithm = HS.HarmonySearch(
#     problem = problem,
#     max_nfes = 10000,
#     harmony_memory_size=100,
#     memory_consider_rate = 0.9,
#     pitch_adjust_rate = 0.5,
#     band_width = 0.5
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)

# algorithm = ICA.ImperialistCompetitiveAlgorithm(
#     problem = problem,
#     max_nfes = 10000,
#     population_size= 100,
#     imperialists_size = 10,
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()

# algorithm = JADE.JADE(
#     problem = problem,
#     max_nfes = 10000,
#     population_size= 30,
#     archive_size = 0,
#     p = 0.05,
#     c = 0.1
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()

# algorithm = WOA.WhaleOptimizer(
#     problem = problem,
#     max_nfes = 10000,
#     swarm_size= 100,
# )
# algorithm.run()
# print("Time:", algorithm.total_computing_time)
# print("Minimum Obj:", algorithm.get_result().objective)
# plt.plot(algorithm.records)
# plt.show()


import numpy as np

class Solution:
    """ Class representing solutions """

    def __init__(self, number_of_variables: int):
        self.number_of_variables = number_of_variables
        self.objective = None
        self.variables = [[] for _ in range(self.number_of_variables)]

    def __eq__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return self.variables == solution.variables
        return False

    def __str__(self) -> str:
        return 'Solution(objective={},variables={})'.format(self.objective, self.variables)

    def __copy__(self):
        new_solution = Solution(self.number_of_variables)
        new_solution.objective = self.objective
        new_solution.variables = self.variables[:]
        return new_solution

class FloatSolution:
    """ Class representing solutions """

    def __init__(self, number_of_variables: int):
        self.number_of_variables = number_of_variables
        self.objective = None
        self.variables = np.zeros(self.number_of_variables)

    def __eq__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return self.variables == solution.variables
        return False

    def __str__(self) -> str:
        return 'Solution(objective={},variables={})'.format(self.objective, self.variables)

    def __copy__(self):
        new_solution = Solution(self.number_of_variables)
        new_solution.objective = self.objective
        new_solution.variables = self.variables.copy()
        return new_solution


'''
Imperialist Competitive Algorithm
@author: zzw
@reference: Atashpaz-Gargari, E., & Lucas, C. (2007).
                    Imperialist competitive algorithm: An algorithm for optimization inspired by imperialistic competition.
                    2007 IEEE Congress on Evolutionary Computation.
                    doi:https://doi.org/10.1109/CEC.2007.4425083

@log: 2019.11.30, creation
'''
from soio.core.algorithm import Algorithm
from soio.core.solution import FloatSolution
from soio.core.problem import FloatProblem
import time
import numpy as np
import random
import math

class ImperialistCompetitiveAlgorithm(Algorithm):

    def __init__(self,
                 problem: FloatProblem,
                 population_size: int,
                 imperialists_size: int,
                 max_nfes: int):

        self.problem = problem
        self.population_size = population_size
        self.imperialists_size = imperialists_size
        self.colonies_size = population_size - imperialists_size
        self.max_infes = max_nfes

        self.beta = 2
        self.gamma = math.pi / 4
        self.epsilon = 0.1


        self.best_solution = FloatSolution(self.problem.number_of_variables)
        self.best_solution.objective = float("inf")

    def create_initial_countries(self):
        """ Creates the initial list of solutions of a metaheuristic. """
        return [self.problem.create_solution() for _ in range(self.population_size)]

    def evaluate(self, countries):
        """ Evaluates the countries. """
        for i in range(self.population_size):
            self.problem.evaluate(countries[i])
        return countries

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        # create initial countries
        self.countries = self.create_initial_countries()
        self.evaluate(self.countries)

        # create initial empires
        self.countries.sort(key=lambda x: x.objective)
        self.best_solution = self.countries[0]
        self.records = []
        imperialists = self.countries[:self.imperialists_size]
        colonies = self.countries[self.imperialists_size:]
        max_imp_cost = imperialists[-1].objective
        normalized_imp_costs = np.array([max_imp_cost-s.objective for s in imperialists])
        imp_powers = normalized_imp_costs / np.sum(normalized_imp_costs)
        imp_colonies_size = [int(round(ip*self.colonies_size)) for ip in imp_powers]
        imp_colonies_size[0] += self.colonies_size - sum(imp_colonies_size)
        assert self.colonies_size == sum(imp_colonies_size), [self.colonies_size,sum(imp_colonies_size) ]
        random.shuffle(colonies)
        imp_colonies  = []
        counter = 0
        for  i in range(self.imperialists_size):
            imp_colonies.append(colonies[counter: counter + imp_colonies_size[i]])
            counter += imp_colonies_size[i]

        while True:
            if self.problem.nfes >= self.max_infes:
                break

            # assimilating: move the colonies toward their relevant imperialist
            for i in range(self.imperialists_size):
                for j in range(imp_colonies_size[i]):
                    colony = imp_colonies[i][j]
                    for k in range(self.problem.number_of_variables):
                        colony.variables[k] = colony.variables[k] + random.random() * self.beta * (imperialists[i].variables[k] - colony.variables[k])
                        colony.variables[k] = max(min(colony.variables[k], self.problem.upper_bound[k]), self.problem.lower_bound[k])
                    self.problem.evaluate(colony)
                    if colony.objective < imperialists[i].objective:
                        # swap colony and imperialist
                        imp_colonies[i][j] = imperialists[i]
                        imperialists[i] = colony

            for imp in imperialists:
                if imp.objective < self.best_solution.objective:
                    self.best_solution = imp

            self.records.append(self.best_solution.objective)

            if self.imperialists_size > 1:
                # imperialistic competition, eliminate the powerless empire
                imp_total_costs = [imperialists[i].objective + self.epsilon *
                                   sum([imp_colonies[i][j].objective for j in range(imp_colonies_size[i])])
                                   for i in range(self.imperialists_size)]
                weakest_imp = np.argmin(imp_total_costs)
                imp_total_costs.pop(weakest_imp)
                imp_colonies_size.pop(weakest_imp)
                new_colonies = imp_colonies.pop(weakest_imp) + [imperialists.pop(weakest_imp),]
                self.imperialists_size -= 1
                norm_imp_total_costs = np.array(imp_total_costs) - max(imp_total_costs)
                imp_powers = norm_imp_total_costs / (np.sum(norm_imp_total_costs) + 1e-9)
                total_power = np.sum(imp_powers)
                for colony in new_colonies:
                    # roulette wheel selection
                    r = random.random() * total_power
                    select_imp = -1
                    while r >0:
                        select_imp += 1
                        r -= imp_powers[select_imp]
                    imp_colonies[select_imp].append(colony)
                    imp_colonies_size[select_imp] += 1

        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Imperialist Competitive Algorithm'




from soio.core.problem import FloatProblem
from soio.core.solution import FloatSolution
import matplotlib.pyplot as plt

class FuncProblem(FloatProblem):
    def __init__(self, func, n_varaibles, lower_bound, upper_bound):
        self.number_of_variables = n_varaibles
        self.nfes = 0
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.func = func

    def evaluate(self, solution: FloatSolution):
        solution.objective = self.func(solution.variables)
        self.nfes += 1
        return solution

def imp(func, Optimizer, opt_params, n_variables, lower_bound, upper_bound, plot= False):
    optimizer = Optimizer(
        problem=FuncProblem(func, n_variables, lower_bound, upper_bound),
        **opt_params
    )
    optimizer.run()
    optimal_solution = optimizer.get_result()
    if plot:
        plt.plot(optimizer.records)
        plt.show()
    return optimal_solution.variables, optimal_solution.objective


'''
Whale Optimization Algorithm with cellular topological structure
@author: zzw
@reference: WOA & CGWO

@ log: 2019.7.25, create
'''
from soio.core.algorithm import Algorithm
from soio.core.problem import FloatProblem
from soio.core.solution import FloatSolution
import random
import time
import numpy as np
import math

'''
Cellular Automata (2D)
'''
CA_L5 = [[0,-1], [0,1], [-1,0], [1,0]]
CA_L9 = [[0,-2], [0,-1], [0,1], [0,2], [-2,0], [-1,0], [2,0], [1,0]]
CA_C9 = [[-1,-1], [0,-1], [1,-1], [-1,0], [1,0], [-1,1], [0,1], [1,1]]
CA_C13 = [[0,-2], [-1,-1], [0,-1], [1,-1], [-2,0], [-1,0], [1,0], [2,0],
          [-1,1], [0,1], [1,1], [0, 2]]
CA_C21 = [[-1,-2], [0,-2], [1,-2],
          [-2,-1], [-1,-1], [0,-1], [1,-1], [2,-1],
          [-2,0], [-1,0], [1,0], [2,0],
          [-2,1], [-1,1], [0,1], [1,1], [2,1],
          [-1,2], [0,2], [1,2]]
CA_C25 = [[-2,-2], [-1,-2], [0,-2], [1,-2], [2,-2],
          [-2,-1], [-1,-1], [0,-1], [1,-1], [2,-1],
          [-2,0], [-1,0], [1,0], [2,0],
          [-2,1], [-1,1], [0,1], [1,1], [2,1],
          [-2,2], [-1,2], [0,2], [1,2], [2,2]]

def find_neighborhood(neighbor_structure, grid_catalog, coord):
    '''
    :param neighbor_structure:  2D list, Cellular Automata
    :param grid_catalog: 2D ndarray, wolf index in each celluar grid
    :param coord:  [x,y]
    :return:  neighborhood wolves indexes
    '''
    l,w = grid_catalog.shape
    neighbors = []
    for nc in neighbor_structure:
        nx = coord[0] + nc[0]
        ny = coord[1] + nc[1]

        if nx >= l:
            nx -= l
        elif nx < 0:
            nx += l

        if ny >= w:
            ny -= w
        elif ny < 0:
            ny += w

        neighbors.append(grid_catalog[nx,ny])
    return neighbors


class CellularWhaleOptimizer(Algorithm):
    def __init__(self,
                 problem: FloatProblem,
                 max_nfes: int,
                 grid_shape = [20,20],
                 neighbor_structure = CA_C21,
                 ):

        self.problem = problem
        self.swarm_size = grid_shape[0]*grid_shape[1]
        self.max_iterations = int(max_nfes/(grid_shape[0]*grid_shape[1]))

        self.neighbor_structure = neighbor_structure
        self.grid_catalog = np.random.permutation(self.swarm_size).reshape(grid_shape)

        self.whales_coords = [None]*self.swarm_size
        for i in range(grid_shape[0]):
            for j in range(grid_shape[1]):
                self.whales_coords[self.grid_catalog[i][j]] = [i,j]

        self.best_solution = FloatSolution(self.problem.number_of_variables)
        self.best_solution.objective = float("inf")


    def create_initial_swarm(self) -> np.ndarray:
        """ Creates the initial list of solutions of a metaheuristic. """
        return [self.problem.create_solution() for _ in range(self.swarm_size)]

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        self.swarm = self.create_initial_swarm()
        for i in range(self.swarm_size):
            self.problem.evaluate(self.swarm[i])
            if self.swarm[i].objective < self.best_solution.objective:
                self.best_solution = self.swarm[i]

        self.records = []

        for iter in range(self.max_iterations):
            a = 2 - iter * ((2) / self.max_iterations)  # a decreases linearly fro 2 to 0
            a2 = -1 + iter * ((-1) / self.max_iterations)  # a2 linearly decreases from -1 to -2

            for i in range(self.swarm_size):
                coord = self.whales_coords[i]
                neighbors = find_neighborhood(self.neighbor_structure, self.grid_catalog, coord)
                leader = self.select_best_neigbor(neighbors)

                r1 = random.random()
                r2 = random.random()
                A = 2 * a * r1 - a
                C = 2 * r2
                b = 1
                l = (a2 - 1) * random.random() + 1
                p = random.random()

                new_whale = FloatSolution(self.problem.number_of_variables)

                for j in range(0, self.problem.number_of_variables):

                    if p < 0.5:
                        if abs(A) >= 1:
                            # rand_leader_index = random.randint(0, self.swarm_size-1)
                            rand_leader_index = random.choice(neighbors)
                            X_rand = self.swarm[rand_leader_index].variables
                            D_X_rand = abs(C * X_rand[j] - self.swarm[i].variables[j])
                            new_whale.variables[j] = X_rand[j] - A * D_X_rand

                        elif abs(A) < 1:
                            D_Leader = abs(C * leader.variables[j] - self.swarm[i].variables[j])
                            new_whale.variables[j] = leader.variables[j] - A * D_Leader

                    elif p >= 0.5:
                        distance2Leader = abs(leader.variables[j] - self.swarm[i].variables[j])
                        new_whale.variables[j] = distance2Leader * math.exp(b * l) * math.cos(l * 2 * math.pi) + leader.variables[j]

                new_whale.variables = np.clip(new_whale.variables, self.problem.lower_bound, self.problem.upper_bound)
                self.problem.evaluate(new_whale)
                # greedy
                if new_whale.objective < self.swarm[i].objective:
                    self.swarm[i] = new_whale
                    if new_whale.objective < self.best_solution.objective:
                        self.best_solution = new_whale

            self.records.append(self.best_solution.objective)

        self.total_computing_time = time.time() - start_computing_time

    def select_best_neigbor(self, neighbors):
        leader = FloatSolution(self.problem.number_of_variables)
        leader.objective = float("inf")
        for i in neighbors:
            if self.swarm[i].objective < leader.objective:
                leader = self.swarm[i]

        return leader



    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Cellular Whale Optimization Algorithm'



'''
Differential Evolution Algorithm
@author: zzw
@reference: R. Storn, K. Price,
                    Differential Evolution A Simple and Efficient Heuristic for global Optimization over Continuous Spaces,
                    Journal of Global Optimization, 11 (1997) 341-359.

@log: 2020.1.3, create
'''

from soio.core.algorithm import Algorithm
from soio.core.solution import FloatSolution
from soio.core.problem import FloatProblem
import time
import random
import copy

class DifferentialEvolution(Algorithm):

    def __init__(self,
                 problem: FloatProblem,
                 population_size: int,
                 max_nfes: int,
                 CR:float = 0.5, # [0,1] crossover rate
                 F: float = 0.1, #[0,2}
                 ):

        self.problem = problem
        self.population_size = population_size
        self.max_iterations = int(max_nfes/population_size)
        self.F = F
        self.CR = CR

        self.best_solution = FloatSolution(self.problem.number_of_variables)
        self.best_solution.objective = float("inf")

    def create_initial_population(self):
        return [self.problem.create_solution() for _ in range(self.population_size)]

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        self.population = self.create_initial_population()
        for ind in self.population:
            self.problem.evaluate(ind)
            if ind.objective < self.best_solution.objective:
                self.best_solution = ind
        self.records = []

        for iter in range(self.max_iterations):
            for i in range(self.population_size):
                a = int(random.random()*self.population_size)
                while i==a:
                    a = int(random.random() * self.population_size)
                b = int(random.random() * self.population_size)
                while i==b or a==b:
                    b = int(random.random() * self.population_size)
                c = int(random.random() * self.population_size)
                while i==c or a==c or b==c:
                    c = int(random.random() * self.population_size)
                new_solution = FloatSolution(self.problem.number_of_variables)
                for j in range(self.problem.number_of_variables):
                    if random.random() < self.CR:
                        v = self.population[a].variables[j] + self.F * (self.population[b].variables[j] - self.population[c].variables[j])
                        new_solution.variables[j] = min(max(v, self.problem.lower_bound[j]), self.problem.upper_bound[j])
                    else:
                        new_solution.variables[j] = self.population[i].variables[j]
                rj = int(random.random()*self.problem.number_of_variables)
                new_solution.variables[rj] = self.population[a].variables[rj] + self.F * (self.population[b].variables[rj] - self.population[c].variables[rj])
                self.problem.evaluate(new_solution)

                if new_solution.objective < self.population[i].objective:
                    self.population[i] = new_solution
                    if new_solution.objective < self.best_solution.objective:
                        self.best_solution = new_solution

            self.records.append(self.best_solution.objective)

        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'Differential Evolution Algorithm'


'''
Differential Evolution Algorithm
@author: zzw
@reference: J. Zhang, A.C. Sanderson,
                    JADE: Adaptive Differential Evolution With Optional External Archive,
                    IEEE Transactions on Evolutionary Computation, 13 (2009) 945-958.


@log: 2020.1.3, create
'''

from soio.core.algorithm import Algorithm
from soio.core.solution import FloatSolution
from soio.core.problem import FloatProblem
import time
import random
from scipy.stats import cauchy
import numpy as np

class JADE(Algorithm):

    def __init__(self,
                 problem: FloatProblem,
                 max_nfes: int,
                 population_size: int = 30,
                 archive_size = 0,
                 p = 0.05, # [5%, 20%]
                 c = 0.1 #[1/20, 1/5]
                 ):

        self.problem = problem
        self.population_size = population_size
        self.max_iterations = int(max_nfes/population_size)
        self.archive_size = archive_size
        self.p = p
        self.c = c

        self.best_solution = FloatSolution(self.problem.number_of_variables)
        self.best_solution.objective = float("inf")

    def create_initial_population(self):
        return [self.problem.create_solution() for _ in range(self.population_size)]

    def run(self):
        """ Execute the algorithm. """
        start_computing_time = time.time()

        self.population = self.create_initial_population()
        for ind in self.population:
            self.problem.evaluate(ind)

        top_population = sorted(self.population, key=lambda x: x.objective)[:int(self.population_size*self.p)]
        archive = []
        mu_CR = 0.5
        mu_F = 0.5

        self.best_solution = top_population[0]
        self.records = []

        for iter in range(self.max_iterations):
            SCR = [] # successful crossover probabilities
            SF =[] # successful mutation factors
            for i in range(self.population_size):
                CR_i = random.normalvariate(mu_F, 0.1)
                F_i = cauchy.rvs(loc=mu_F, scale=0.1, size=1)[0]
                Xp_best = top_population[int(random.random()*len(top_population))]

                a = int(random.random()*self.population_size)
                while i==a:
                    a = int(random.random() * self.population_size)
                b = int(random.random() * (self.population_size+len(archive)))
                while i==b or a==b:
                    b = int(random.random() * self.population_size+len(archive))
                Xi = self.population[i]
                Xa = self.population[a]
                Xb = self.population[b] if b < self.population_size else archive[b-self.population_size]

                new_solution = FloatSolution(self.problem.number_of_variables)
                for j in range(self.problem.number_of_variables):
                    if random.random() < CR_i:
                        v = Xi.variables[j] + F_i * (Xp_best.variables[j] - Xi.variables[j]) + F_i*(Xa.variables[j] - Xb.variables[j])
                        #new_solution.variables[j] = min(max(v, self.problem.lower_bound[j]), self.problem.upper_bound[j])
                        if v < self.problem.lower_bound[j]:
                            v = (self.problem.lower_bound[j] + Xi.variables[j]) / 2
                        if v > self.problem.upper_bound[j]:
                            v = (self.problem.upper_bound[j] + Xi.variables[j]) / 2
                        new_solution.variables[j] = v
                    else:
                        new_solution.variables[j] = Xi.variables[j]
                rj = int(random.random()*self.problem.number_of_variables)
                v =Xi.variables[rj] + F_i * (Xp_best.variables[rj] - Xi.variables[rj]) + F_i*(Xa.variables[rj] - Xb.variables[rj])
                if v < self.problem.lower_bound[rj]:
                    v = (self.problem.lower_bound[rj] + Xi.variables[rj]) / 2
                if v > self.problem.upper_bound[rj]:
                    v = (self.problem.upper_bound[rj] + Xi.variables[rj]) / 2
                new_solution.variables[rj] = v
                self.problem.evaluate(new_solution)

                if new_solution.objective < self.population[i].objective:
                    if self.archive_size > 0:
                        if len(archive)<self.archive_size:
                            archive.append(self.population[i])
                        else:
                            archive[int(random.random()*self.archive_size)] = self.population[i]
                    self.population[i] = new_solution
                    SCR.append(CR_i)
                    SF.append(F_i)

            if len(SCR) > 0:
                mu_CR = (1-self.c)*mu_CR + self.c * np.mean(SCR)
                mu_F = (1-self.c) * mu_F + self.c * np.sum(np.power(SF,2)) / (np.sum(SF)+1E-10)

            top_population = sorted(self.population, key=lambda x: x.objective)[:int(self.population_size * self.p)]
            self.best_solution = top_population[0]
            self.records.append(self.best_solution.objective)

        self.total_computing_time = time.time() - start_computing_time

    def get_result(self):
        return self.best_solution

    def get_name(self) -> str:
        return 'JA Differential Evolution Algorithm'